//##########################################################
//
// Project : Oddsnake version 1.15
//
// Author : Cyril MOTTIER
//
// Mail : orlin61@wanadoo.fr
//
// Site : www.cyrilmottier.com
//
// C Source File (main program for v200 and 92p)
//
// Created 2004/04/9 with TIGCC 0.95 beta 5 10:37:06
//
//###########################################################

// Comments on this source for everyone who wants to start learning C on Ti68k
// This source will show you how to use TIGCC functions. It's easy to understand because
// no complex algorythm are used. Some parts of this source can be easily optimized but
// I dont' want to do this optimization because it may change all the game and the maniability
// For exemple, the algorythme to rotate the array which contains infos of the queue are not the faster
// You can use another method (named FIFO) which use a integer. This integer indicates the head and 
// rotate using the modulus operator. You can also optimize the maniability by puting the 
// key-tests in the waiting loop. Nevertheless it will totaly change the speed of this game.

#define USE_TI92PLUS
#define USE_V200  //Produce a 92+ and v200 file
#define OPTIMIZE_ROM_CALLS
#include <tigcclib.h>
#include "sprites_def.h"
#define MAGIC 187 // This number must be odd on V200 and Ti92+ and even on Ti89 and Ti89t
#define HW ((HW_VERSION==1) ? 5 : 6)


//Structure for snake tail
typedef struct
{
  int x_tail;
  int y_tail;
  int orientation; // 0:left # 1:up # 2:right # 3:down
} TAIL;

//Structure for highscores
typedef struct
{
  char name[9];
  int score;
} HSC_ITEM;

typedef struct
{
	int x;
	int y;
} COORD;

COORD *ptr_pos;
COORD food={84,59};
TAIL *ptr_past;

// Initial highscores table 
static HSC_ITEM bestscore[5]={{ "Cyril",100}, {"Cyril",100}, {"Cyril",100}, { "Cyril",100}, {"Cyril",100}};

//Variables for level, mazes choice and food
int choice[2]={1,1};

short _title(void)
{
  short i, a=68, add=3, line[6]={60,61,63,180,179,177};
  const char *info[5]={"Oddsnake v1.15", "Coded by Cyril MOTTIER", "on November 2005", "orlin61@wanadoo.fr", "www.cyrilmottier.com"};
  const char *titres[5]={"Game", "Options", "Highscores", "About", "Quit"};
 
  mode:
  ClrScr ();
  BitmapPut (50,10,(BITMAP *)&(unsigned char[])
{0,40,0,140,0x01,0xF8,0x00,0x1E,0x00,0x0F,0x00,0x0F,0x00,0x20,0x00,0x00,0x70,0x18,0x00,0x00,0x0E,0x00,
0x0F,0xFE,0x00,0x3F,0x00,0x1F,0x80,0xFF,0xC1,0xF0,0x00,0x07,0xFC,0x3F,0x80,0x00,0xFF,0x80,
0x0F,0xFE,0x00,0x3F,0x00,0x1F,0x80,0xFF,0xC1,0xF0,0x00,0x07,0xFC,0x3F,0x80,0x00,0xFF,0x80,
0x1F,0xFE,0x00,0x3F,0x80,0x1F,0xC3,0xFF,0xC3,0xF0,0xF0,0x1F,0xFC,0x3F,0xC0,0x00,0xFF,0xC0,
0x1F,0xFE,0x00,0x3F,0x80,0x1F,0xC3,0xFF,0xE3,0xFF,0xF8,0x7F,0xFC,0x3F,0xC0,0x01,0xFF,0xE0,
0x1F,0xFE,0x00,0x3F,0x80,0x1F,0xC3,0xFF,0xE3,0xFF,0xF8,0x7F,0xFC,0x3F,0xC0,0x01,0xFF,0xE0,
0x3F,0xFF,0x00,0x3F,0x80,0x1F,0xC7,0xFF,0xE3,0xFF,0xF8,0x7F,0xFC,0x3F,0xC0,0x61,0xFF,0xE0,
0x3F,0xFF,0x00,0x1F,0x80,0x0F,0xC7,0xFF,0xE3,0xFF,0xFC,0x7F,0x7C,0x7F,0xC0,0xF1,0xFF,0xE0,
0x3F,0xFF,0x00,0x1F,0x80,0x0F,0xC7,0xFF,0xE3,0xFF,0xFC,0x7F,0x7C,0x7F,0xC0,0xF1,0xFF,0xE0,
0x3F,0x9F,0x00,0x1F,0x80,0x0F,0xC7,0xF7,0xF3,0xFF,0xFC,0xFE,0x3E,0x7F,0xC1,0xF9,0xFF,0xE0,
0x3F,0x9F,0x00,0x1F,0x80,0x0F,0xC7,0xE3,0xF3,0xFF,0xFC,0xFE,0x3E,0x7F,0xC3,0xF3,0xE3,0xF0,
0x3F,0x9F,0x00,0x1F,0x80,0x0F,0xC7,0xE3,0xF3,0xFF,0xFC,0xFE,0x3E,0x7F,0xC3,0xF3,0xE3,0xF0,
0x3F,0x9F,0x00,0x1F,0xC0,0x0F,0xE7,0xE3,0xE3,0xF9,0xFC,0xFE,0x3E,0x7F,0xC7,0xF3,0xE3,0xF0,
0x3F,0x9F,0x87,0xFF,0xC1,0xFF,0xE7,0xF1,0x83,0xF9,0xFC,0xFC,0x3E,0x7F,0xCF,0xE3,0xE3,0xF0,
0x3F,0x9F,0x87,0xFF,0xC1,0xFF,0xE7,0xF1,0x83,0xF9,0xFC,0xFC,0x3E,0x7F,0xCF,0xE3,0xE3,0xF0,
0x3F,0x8F,0x8F,0xFF,0xC3,0xFF,0xE3,0xF0,0x03,0xF9,0xFC,0x7C,0x3E,0x7F,0x9F,0xE3,0xE3,0xF0,
0x3F,0x8F,0x8F,0xFF,0xC3,0xFF,0xE3,0xF0,0x03,0xF9,0xFE,0x00,0x3E,0x7F,0xFF,0xC3,0xE3,0xF0,
0x3F,0x8F,0x8F,0xFF,0xC3,0xFF,0xE3,0xF0,0x03,0xF9,0xFE,0x00,0x3E,0x7F,0xFF,0xC3,0xE3,0xF0,
0x3F,0x8F,0x9F,0xFF,0xC7,0xFF,0xE3,0xFF,0xE3,0xF9,0xFE,0x00,0x3E,0x7F,0xFF,0xC3,0xEF,0xE0,
0x3F,0x8F,0x9F,0xFF,0xC7,0xFF,0xE3,0xFF,0xFB,0xF9,0xFE,0x07,0xFE,0x7F,0xFF,0x83,0xFF,0xE0,
0x3F,0x8F,0x9F,0xFF,0xC7,0xFF,0xE3,0xFF,0xFB,0xF9,0xFE,0x07,0xFE,0x7F,0xFF,0x83,0xFF,0xE0,
0x7F,0x0F,0x9F,0xCF,0xC7,0xE7,0xE3,0xFF,0xFB,0xF9,0xFE,0x1F,0xFE,0x7F,0xFF,0x83,0xFF,0xE0,
0x7F,0x0F,0xDF,0xCF,0xC7,0xE7,0xE0,0xFF,0xFB,0xF9,0xFE,0x3F,0xFE,0x7F,0xFF,0x03,0xFF,0xE0,
0x7F,0x0F,0xDF,0xCF,0xC7,0xE7,0xE0,0xFF,0xFB,0xF9,0xFE,0x3F,0xFE,0x7F,0xFF,0x03,0xFF,0xE0,
0x7F,0x0F,0xDF,0x8F,0xE7,0xC7,0xF0,0x07,0xFF,0xF9,0xFE,0x7F,0x3F,0x7F,0xFF,0x03,0xFC,0x00,
0x7F,0x0F,0xDF,0x8F,0xE7,0xC7,0xF0,0x03,0xFF,0xF9,0xFE,0x7F,0x3F,0x7F,0xFF,0x03,0xF1,0xE0,
0x7F,0x0F,0xDF,0x8F,0xE7,0xC7,0xF0,0x03,0xFF,0xF9,0xFE,0x7F,0x3F,0x7F,0xFF,0x03,0xF1,0xE0,
0x7F,0x0F,0xDF,0x8F,0xE7,0xC7,0xF1,0xE1,0xFF,0xF8,0xFE,0x7F,0x3F,0x7F,0xFF,0x03,0xF7,0xE0,
0x7F,0xFF,0x9F,0x8F,0xE7,0xC7,0xF3,0xE1,0xFF,0xF8,0xFE,0x7E,0x3F,0xFF,0xFF,0x83,0xE7,0xF0,
0x7F,0xFF,0x9F,0x8F,0xE7,0xC7,0xF3,0xE1,0xFF,0xF8,0xFE,0x7E,0x3F,0xFF,0xFF,0x83,0xE7,0xF0,
0x7F,0xFF,0x9F,0xCF,0xE7,0xE7,0xF3,0xE3,0xFF,0xF8,0xFF,0x7E,0x3F,0xFF,0x7F,0xC7,0xE7,0xF0,
0x7F,0xFF,0x3F,0xFF,0xEF,0xFF,0xF3,0xFF,0xFF,0xF8,0xFF,0xFF,0x7E,0xFF,0x3F,0xE7,0xFF,0xF0,
0x7F,0xFF,0x3F,0xFF,0xEF,0xFF,0xF3,0xFF,0xFF,0xF8,0xFF,0xFF,0x7E,0xFF,0x3F,0xE7,0xFF,0xF0,
0x7F,0xFE,0x3F,0xFF,0xEF,0xFF,0xF3,0xFF,0xFF,0xF8,0xFF,0xFF,0xFE,0xFF,0x1F,0xE7,0xFF,0xF0,
0x3F,0xFE,0x3F,0xFF,0xCF,0xFF,0xE3,0xFF,0xFB,0xF8,0xFF,0xFF,0xFE,0xFF,0x1F,0xC3,0xFF,0xF0,
0x3F,0xFE,0x3F,0xFF,0xCF,0xFF,0xE3,0xFF,0xFB,0xF8,0xFF,0xFF,0xFE,0xFF,0x1F,0xC3,0xFF,0xF0,
0x1F,0xFC,0x3F,0xFF,0x8F,0xFF,0xC3,0xFF,0xFB,0xF0,0xFF,0x7F,0xFC,0xFF,0x0F,0x83,0xFF,0xE0,
0x03,0xFC,0x1F,0xF8,0x07,0xFC,0x01,0xFF,0xF3,0xF0,0xFE,0x7F,0xF0,0x7E,0x07,0x03,0xFF,0xE0,
0x03,0xFC,0x1F,0xF8,0x07,0xFC,0x01,0xFF,0xF3,0xF0,0xFE,0x7F,0xF0,0x7E,0x07,0x03,0xFF,0xE0,
0x00,0x70,0x0E,0x00,0x03,0x80,0x01,0xFF,0xC1,0xE0,0x78,0x1C,0x00,0x1C,0x03,0x00,0x0F,0x80},&(SCR_RECT){{0,0,239,127}},A_REPLACE);
  FontSetSys (F_6x8);
  short key=338;
	  for (i=0; i<5; i++) DrawStr (120-strlen(titres[i])*3, i*10+70, titres[i], A_NORMAL);
	  for(i=0; i<6; i++) DrawLine (line[i], 67, line[i], 119, A_NORMAL);
    while (key!=264)
	  {
	     if (key==338||key==344) ScrRectFill (&(SCR_RECT){{66,a,174,a+10}}, &(SCR_RECT){{0,0,239,127}}, A_XOR);
	     key=ngetchx();
		   if (key==338||key==344) ScrRectFill (&(SCR_RECT){{66,a,174,a+10}}, &(SCR_RECT){{0,0,239,127}}, A_XOR);
	     if (key==338)
	     {
	  	   a-=10;
	  	   if(a<68) a=108;
	     }
	     if (key==344)
	     {
	       a+=10;
	       if (a>108) a=68;
	     }
	     if (key==13)
	     {
	       if (a==68) return 1;
	       if (a==108) return 0;
	       ScrRectFill(&(SCR_RECT){{59,67,181,119}}, &(SCR_RECT){{0,0,239,127}}, A_REVERSE);
	       //If you are on 'Options'
	  	     if (a==78)
	  	     {
             key=0;
	  	       while(key!=264&&key!=13)
	  	       {
	  	         printf_xy (54, 85, "[1] : Choose speed : %d", choice[0]);
	  	         printf_xy (56, 95, "[2] : Choose maze : %d", choice[1]);
	  	         key=ngetchx();
	  	   	     if (key==49)
	  	   	     {
	  	   	       choice[0]++;
	  	           if (choice[0]>5) choice[0]=1;
	  	   	   	 }
	  	   	     if (key==50)
	  	   	     {
	  	   	 	     choice[1]++;
	  	   	 	     if (choice[1]>5) choice[1]=1;
	  	   	 	   }
	  	       }
	         }
	         //If you are on 'highscores'
	         if (a==88)
	         {
	      	   //Display your highscores (maybe...)
	           for (i=0;i<5;i++) 
	     	     {
	     		     if (bestscore[i].score>1000) add=4;
	     	       printf_xy (120-(strlen(bestscore[i].name)+4+add)*3, i*10+70, "%d by %s", bestscore[i].score, bestscore[i].name); 
	     	       add=3;
	     	     }
        	   ngetchx();
	         }
	  	     //If you are on 'About'
	  	     if (a==98)	
	         {
	  	     //Display information about me...
	  	       for (i=0;i<5;i++) DrawStr(120-strlen(info[i])*3, i*10+70, info[i], A_NORMAL);
	  	       ngetchx ();
	  	     }
	  		   goto mode;	
	       }	
       }
  return 0;
}

//Function to draw a dialog box
void _box ()
{
  short c, w;
  for (c=0; c<12; c++)
	 {
		DrawLine (20, 64-c, 220, 64-c, A_NORMAL);
		DrawLine (21, 65-c, 219, 65-c, A_REVERSE);
		 for (w=0; w<8000; w++);
		DrawLine (20, 65+c, 220, 65+c, A_NORMAL);
		DrawLine (21, 64+c, 219, 64+c, A_REVERSE);
		}
}

/*===========================================================================*/
/* "waits" a given number of milliseconds (dummy looping)                    */
/* NOTE: this wait loop is parametrized to fit on HW1 calculators            */
/*===========================================================================*/
//Thanks to Thomas Nussbaumer for this routine extracted from Ti-Chess source
void WaitForMillis(unsigned short);

asm("xdef WaitForMillis\n"
"WaitForMillis:  movem.l %d0-%d3,-(%sp)\n"
"           move.w (20,%sp),%d2\n"
"           move.l #31,%d1\n"
"           move.l #31,%d3\n"
"_wl2_:     move.w #120,%d0    /* modify this value for exact timing !!! */\n"
"_wl1_:     rol.l  %d3,%d1\n"
"           dbra   %d0,_wl1_\n"
"           dbra   %d2,_wl2_\n"
"           movem.l (%sp)+,%d0-%d3\n"
"           rts");


//Function to draw  a sprite but this sprite replace the background
void Rpl_Sprite(int x, int y, char *sprite)
{
 Sprite8 (x, y, 8, snake_sprt[1], LCD_MEM, SPRT_AND);
 Sprite8 (x, y, 8, sprite, LCD_MEM, SPRT_OR);
}

//Function to input your name if you have an highscore
void input_name(int score, int mode)
{
  _box();
  FontSetSys (F_8x10);
  if (mode==1) DrawStr (76, 60, "You lose!!!", A_NORMAL);
  else DrawStr (80, 60, "You win!!!", A_NORMAL);
  while (ngetchx()!=13);
  short a, i=0,key=0;
	  if (score>bestscore[4].score)
	  {
    for(a=4;(bestscore[a-1].score<score)&&a;a--) memcpy(&bestscore[a],&bestscore[a-1],sizeof(HSC_ITEM));
    bestscore[a].score=score;
    
    FontSetSys (F_6x8);		
    //open the box
    _box();
    DrawStr (28, 56, "New highscore! Enter your name:", A_NORMAL);
	    while (key!=13&&key!=264)
	    {
	    	 if (i==8) bestscore[a].name[i]=0;
		     else 
		     {
		      bestscore[a].name[i]='_';
		      bestscore[a].name[i+1]=0;
		     }
	     //draw a white rectangle to delete the old name
	     ScrRectFill (&(SCR_RECT){{21, 66, 219, 74}}, &(SCR_RECT){{0, 0, 239, 127}}, A_REVERSE);
      //draw your name in the center of the screen
	     DrawStr (120-strlen (bestscore[a].name)*3, 66, bestscore[a].name, A_NORMAL);
	     key=ngetchx();
	     //check if the char can be printed...
	      if (key>=' '&&key<='~'&&i<8) bestscore[a].name[i++]=key;
	     //check if you want to erase your last letter
	      if(key==257&&i) i--;
     }
	  //and finish the string with 0 !
	  bestscore[a].name[i]=0;
	   if (bestscore[a].name[0]==0) strcpy (bestscore[a].name, "???");
	  }
}

//Thanx to David Coz :
//Functions write and open are extracted from his jezzball game
//Nevertheless, these functions are slightly modified
//Write infos in the file 'oddsave'
void write()
{
FILE *fp;

fp = fopen("oddsave","wb"); //Open a file wb => Write in Binary Mode
fputc(MAGIC,fp); //Write the 'Magic Number'
fwrite(bestscore,sizeof(bestscore),1,fp); // Write highscore;
fwrite(choice,sizeof(choice),1,fp); //Write choice 

// Put extension
fputc(0,fp);
fputs("HSC",fp);
fputc(0,fp);
fputc(OTH_TAG,fp);

fclose(fp);//Close the File
EM_moveSymToExtMem (SYMSTR ("oddsave") ,HS_NULL);	// archive the file
}


void open()
{
FILE *fp;

EM_moveSymFromExtMem (SYMSTR ("oddsave") ,HS_NULL); // unarchive the file if it necessary
fp = fopen("oddsave","rb"); // Open a File rb => Read in Binary mode
if (fp != NULL && (fgetc(fp)==MAGIC))
{
  fread(bestscore,sizeof(bestscore),1,fp);
  fread(choice,sizeof(choice),1,fp);
}
// if the file doesn't exist or is a fake (create/overwrite it)
// It also overwrite a highscore file, if this one have been created 
// on another model of calculator because the magic and the version nulmbers will be different
else fp = fopen("oddsave", "wb");
fclose(fp);
}

//Function to place the food and check if you're wining (return 1)
int place_food(void)
{
  //This function starts to find all the free 'cells' and then
  //put its coordinates in an array
  int a, b, c=0;
  for (b=11;b<122;b+=8)
   {
   	for (a=4;a<233;a+=8) 
   	 {
      if (!(GetPix (a+3, b)))
   	 	 {
   		   ptr_pos[c].x=a;
   	     ptr_pos[c++].y=b;
   		 }
     }
   }
  if (c==0) return 1; // You win
  food = ptr_pos[random(c)];
  Sprite8 (food.x, food.y, 8, snake_sprt[2], LCD_MEM, SPRT_OR);
  return 0;
}

//Function to draw a maze
void draw_maze()
{
 int i,j;
 Sprite32 (4, 11, 14, mazes[choice[1]-2], LCD_MEM, SPRT_OR);
	for (i=29;i>-1;i--)
	 {
		 for (j=14;j>-1;j--)
		  {
		   if ( GetPix(i+4, j+11)) Rpl_Sprite (i*8+4, j*8+11, snake_sprt[11]);
		   else Rpl_Sprite (i*8+4, j*8+11, snake_sprt[1]);
	    }
	 }
}

//Function to draw the game set and place a maze depending on 'Options'
void _drawset()
{
	short i;
	ClrScr();
	 if (choice[1]>1) draw_maze();
 DrawClipRect (&(WIN_RECT){2, 9,237, 124}, &(SCR_RECT){{0,0,239,127}}, A_NORMAL);
 DrawStr (5, 1, "Score :", A_NORMAL);
	Sprite8 (76 , 59, 8, snake_sprt[9], LCD_MEM, SPRT_OR);
 Sprite8 (148 , 59, 8, snake_sprt[5], LCD_MEM, SPRT_OR);
	 for (i=84;i<141;i+=8) Sprite8 (i, 59, 8, snake_sprt[0], LCD_MEM, SPRT_OR);
	place_food();
}


// Main Function
void _main(void)
{
 ptr_past = calloc(406,sizeof(TAIL));
 ptr_pos = calloc(406,sizeof(COORD));
 open();
  
 restart:
 if (_title())
 { 
   OSSetSR(0x0400);
   int x_graph=148, y_graph=59, i, dir=2,length=10, score=0, decrease=0;
 
   //For-loop to put right numbers in TAIL struture when you start 
   for (i=0;i<length;i++)
   {
     ptr_past[i].x_tail=x_graph-(i<<3);
     ptr_past[i].y_tail=59;
     ptr_past[i].orientation=2;	
   }
   _drawset();
	   
   //Variable for snake direction ex:'Right':x_move=1 and y_move=0
   signed char x_move=1, y_move=0;
   
   while(!(_keytest (RR_ESC)))
   {
     if (_keytest (RR_LEFT) && x_move!=1) { x_move=-1;y_move=0;dir=0; }
     if (_keytest (RR_UP) && y_move!=1) { x_move=0;y_move=-1;dir=1; }
     if (_keytest (RR_RIGHT) && x_move!=-1) { x_move=1;y_move=0;dir=2; }
     if (_keytest (RR_DOWN) && y_move!=-1) { x_move=0;y_move=1;dir=3; }
     if (_keytest (RR_CLEAR)) off();
	      
     //Updates snake's position
     x_graph+=x_move<<3;
     y_graph+=y_move<<3; 
     if (x_graph>228) x_graph=4;
     if (x_graph<4) x_graph=228;
     if (y_graph>115) y_graph=11;
     if (y_graph<11) y_graph=115;
	  	 
     //Erases snake's tail using a white sprite
     Sprite8 (ptr_past[length-1].x_tail, ptr_past[length-1].y_tail, 8, snake_sprt[1], LCD_MEM, SPRT_AND);
	  	 
     //This routine checks if the snake can stand where it wants to go so if you lose
     if (GetPix (x_graph+3, y_graph))
     {
       input_name(score, 1); //You lose
       goto restart;
     }
	     
     //Display the snake's head
     Rpl_Sprite (x_graph ,y_graph, snake_sprt[dir+3]);
     //Replace the 'old' head with a body sprite
     Rpl_Sprite (ptr_past[0].x_tail, ptr_past[0].y_tail, snake_sprt[0]);
     //And finally display the tail in its new position
     Rpl_Sprite (ptr_past[length-2].x_tail, ptr_past[length-2].y_tail, snake_sprt[(ptr_past[length-3].orientation)+7]);
		
     //Checks if the snake is eating a bubble
     if(x_graph==food.x&&y_graph==food.y)
     {
       length++;
       decrease=0;
       score=score+((choice[0]+choice[1]-1)<<1);
       if (place_food()) //You win
       {
         input_name(score, 2);
         goto restart;
       }
     }
     decrease++;
     if (decrease==40)
     {
       if (score>0) score--;
       decrease=0;
     } 
     printf_xy (53, 1, "%d  ", score);
	     
     //Rotate the 'TAIL' array
     for (i=length-2;i>-1;i--) memcpy (&ptr_past[i+1], &ptr_past[i], sizeof(TAIL));
	  	
     ptr_past[0].x_tail=x_graph;
     ptr_past[0].y_tail=y_graph;
     ptr_past[0].orientation=dir;
     
     //Please wait!!!
     WaitForMillis((76-(choice[0]<<3))*HW);
	 }
   input_name(score, 1);
   goto restart;
 }
 free (ptr_past);
 free (ptr_pos);
 write();
}    
//Oddsnake By Cyril Mottier (orlin61@wanadoo.fr)